import json
import os
import re

from elsp_env_manager.base.model_manager import load_model


class Variables(object):

    def __init__(self, map_json: dict, define_json: dict, inf=None):
        self.map_json = map_json
        self.define_json = define_json
        self.inf = inf

    @staticmethod
    def find_variable(_str: str):
        c = _str.count('$')
        assert c % 2 == 0
        st = 0
        ed = -1
        v_list = []
        for i in range(c//2):
            st = _str.find("$", ed + 1)
            ed = _str.find("$", st + 1)
            v_list.append(_str[st + 1:ed])
        return v_list

    def get_variable(self, name):
        assert self.inf is not None
        if name in self.map_json:
            map_json = self.map_json[name]
            k = map_json["process_name"] + "_" + map_json["m_id"]
            if k in self.inf and map_json["label"] in self.inf[k]:
                return self.inf[k][map_json["label"]]
            else:
                return 0
                #raise RuntimeWarning
        elif name in self.define_json:
            expression = self.define_json[name]
            if isinstance(expression, str):
                v_list = Variables.find_variable(expression)
                _locals = locals()
                cnt = 0
                for v in v_list:
                    # print(v, self.get_variable(v))
                    v_name = "var_" + str(cnt)
                    cnt += 1
                    _locals[v_name] = self.get_variable(v)
                    expression = expression.replace("${}$".format(v), v_name)
                try:
                    res = eval(expression)
                except KeyError:
                    return 0
                except ZeroDivisionError:
                    print('ZeroDivisionError', name, self.inf)
                    return 0
                except AttributeError:
                    print(name, expression, self.inf)
                    raise AttributeError
                except Exception:
                    print(name, self.inf)
                    raise Exception
                return res
            else:
                raise NotImplementedError
        else:
            print(name)
            raise NotImplementedError
        pass

    @staticmethod
    def load(map_json_file, define_json_file, inf):
        if isinstance(map_json_file, dict):
            map_json = map_json_file
        else:
            map_json = load_model(map_json_file)
        if isinstance(define_json_file, dict):
            define_json = define_json_file
        else:
            define_json = load_model(define_json_file)

        return Variables(map_json=map_json, define_json=define_json, inf=inf)




import math
import numpy as np
from abc import ABC, abstractmethod

from elsp_env_manager.base import Distribution, MID, Substance, Material, NormalDistribution
from elsp_env_manager.base.model_manager import load_model
from elsp_env_manager.base.constants import *


class Process(ABC):
    to_refresh_keys = ['time']

    @abstractmethod
    def __init__(self, process_unit):
        self.process_unit = process_unit
        self.obs_inf_k = self.__class__.__name__.lower() + "_" + str(self.process_unit.m_id)

    def get_attr(self, name, prod: Substance):  # which_may_be_up_to_product
        if getattr(self, name) == "up_to_product":
            return prod.model_json[name]
        else:
            return getattr(self, name)

    def get_distribution(self, name, prod: Substance):
        if isinstance(prod, MID):
            prod = Substance.substance[prod]
        if getattr(self, name) == "up_to_product":
            return Distribution.load(prod.model_json, name)
        else:
            return getattr(self, name)

    def set_obs_inf(self, k, v, replace=True):
        if not hasattr(self.process_unit.env, 'scheduling_unit'):
            return
        if k not in self.process_unit.env.scheduling_unit.inf:
            replace = True
        if replace:
            self.process_unit.env.scheduling_unit.inf[k] = v
        else:
            for v_k, v_v in v.items():
                self.process_unit.env.scheduling_unit.inf[k][v_k] = v_v

    def refresh_obs_inf_keys(self):
        for k in self.to_refresh_keys:
            for obs_inf_k, v in self.process_unit.env.scheduling_unit.inf.items():
                if k in self.process_unit.env.scheduling_unit.inf[obs_inf_k]:
                    self.process_unit.env.scheduling_unit.inf[obs_inf_k].pop(k)

    def get_obs_inf(self, k, sk, default):
        if k not in self.process_unit.env.scheduling_unit.inf:
            self.process_unit.env.scheduling_unit.inf[k] = default
        if sk not in self.process_unit.env.scheduling_unit.inf[k]:
            self.process_unit.env.scheduling_unit.inf[k][sk] = default[sk]
        return self.process_unit.env.scheduling_unit.inf[k]

    def refresh_increment_inf(self, k, v):
        org = self.get_obs_inf(self.obs_inf_k, k, {
            k: 0
        })
        org[k] += v
        self.set_obs_inf(self.obs_inf_k, org)
        pass

    def refresh_inf(self):
        pass

    @abstractmethod
    def reset(self):
        pass


class ProductionProcess(Process):

    def reset(self):
        super().reset()
        self.finished = False
        self.finish_time = 0
        self.cost = 0
        self.last_prod = None
        pass

    def __init__(self, process_unit):
        super().__init__(process_unit)
        self.finished = False
        self.finish_time = -1
        self.start_time = -1
        self.cost = 0
        self.last_prod = None

    def process(self, prod, lead_process=None):
        if lead_process is not None and not lead_process.finished:
            self.finished = False
            self.start_time = -1
            self.set_obs_inf(self.obs_inf_k, {'time':True}, replace=False)
        elif prod is not None and not self.finished:
            t = self.process_unit.get_time()
            if self.start_time == -1:
                self.start_time = t
            lead_process_finish_time = 0 if lead_process is None else lead_process.finish_time
            if self._justify(prod, t - lead_process_finish_time):
                self.finished = True
                self.finish_time = t
                self.last_prod = prod
            self.set_obs_inf(self.obs_inf_k, {'time':{'start':self.start_time, 'current':t,
                                                      'end':self.finish_time,
                                                      'env_time': self.process_unit.env.time,
                                                      'prod': prod.m_id.get_level_value(1)}}, replace=False)
        else:
            self.set_obs_inf(self.obs_inf_k, {'time':True}, replace=False)
            pass
        return self.finished

    @abstractmethod
    def _justify(self, prod, t):
        pass


class Setup(ProductionProcess):

    def __init__(self, process_unit):
        super().__init__(process_unit)
        # self.obs_inf_k = "setup_" + str(self.process_unit.m_id)

    def refresh_inf(self):
        self.refresh_increment_inf("setup cost", self.cost)
        self.cost = 0

    @abstractmethod
    def _justify(self, prod, t):
        pass

    @staticmethod
    def load(json, process_unit):
        if 'setup' not in json:
            return None
        else:
            json = json['setup']
        if json["type"] == "changeover":
            return ChangeoverSetup(setup_expiry=Distribution.load(json, "setup_expiry"),
                                   changeover_time=Distribution.load(json, "changeover_time"),
                                   changeover_cost=Distribution.load(json, "changeover_cost"),
                                   turnaround_time=Distribution.load(json, "turnaround_time"),
                                   turnaround_cost=Distribution.load(json, "turnaround_cost"),
                                   process_unit=process_unit)
        else:
            raise NotImplementedError


class FixedSetup(Setup):

    def __init__(self, time: Distribution, cost: Distribution, process_unit):
        super().__init__(process_unit)
        self.time = time
        self.cost = cost

    def _justify(self, prod, t):
        time = self.time.sample(t)
        finish = t >= time
        if finish:
            self.cost = self.cost.sample(t)
        self.refresh_inf()
        return finish


class ChangeoverSetup(Setup):

    def reset(self):
        super().reset()
        self.changeover_cost_v = 0

    def __init__(self, setup_expiry: Distribution, changeover_time: Distribution, turnaround_time: Distribution,
                 changeover_cost: Distribution,
                 turnaround_cost: Distribution, process_unit):
        super().__init__(process_unit)
        self.setup_expiry = setup_expiry
        self.changeover_time = changeover_time
        self.changeover_cost = changeover_cost
        self.turnaround_time = turnaround_time
        self.turnaround_cost = turnaround_cost

        self.changeover_cost_v = 0

    def needs_changeover(self, prod, t):
        return (not prod.same_series(self.last_prod)) or (
                self.process_unit.process_start_time - self.process_unit.last_process_start_time) >= self.get_distribution(
            "setup_expiry", prod).sample(t)

    def _justify(self, prod: Substance, t):
        time = self.get_distribution("changeover_time", prod).sample(
            t) if self.needs_changeover(prod, t) else self.get_distribution("turnaround_time",
                                                                            prod).sample(t)
        finish = t >= time
        if finish:
            if self.needs_changeover(prod, t):
                changeover_cost = self.get_distribution("changeover_cost", prod).sample(t)
                self.cost = changeover_cost
                self.changeover_cost_v = changeover_cost
                self.refresh_increment_inf("changeover cost", changeover_cost)
            else:
                self.cost = self.get_distribution("turnaround_cost", prod).sample(t)
        self.refresh_inf()
        return finish


class Failure(ProductionProcess):

    def __init__(self, failure_rate: Distribution, failure_cost_per_time: Distribution, process_unit):
        super().__init__(process_unit)
        self.failure_rate = failure_rate
        self.failure_cost_per_time = failure_cost_per_time
        # self.obs_inf_k = 'failure_' + str(self.process_unit.m_id)

    def _justify(self, prod, t):
        rate = self.get_distribution("failure_rate", prod).sample(t)
        failed = np.random.uniform() <= rate
        if failed:
            self.cost = self.get_distribution("failure_cost_per_time", prod).sample(t)
            self.refresh_increment_inf("failure cost", self.cost)
        self.refresh_inf()
        return failed

    @staticmethod
    def load(json, process_unit):
        if 'failure' not in json:
            return None
        else:
            json = json['failure']
        if json["type"] == "failure":
            return Failure(failure_rate=Distribution.load(json, "failure_rate"),
                           failure_cost_per_time=Distribution.load(json, "failure_cost_per_time"),
                           process_unit=process_unit)
        else:
            raise NotImplementedError


class Launch(ProductionProcess):

    def __init__(self, launch_time: Distribution, launch_cost: Distribution, process_unit):
        super().__init__(process_unit)
        self.launch_time = launch_time
        self.launch_cost = launch_cost
        # self.obs_inf_k = "launch_" + str(self.process_unit.m_id)

    def _justify(self, prod: Substance, t):
        time = self.launch_time.sample(t)
        finish = t >= time
        if finish:
            self.cost = self.get_distribution("launch_cost", prod).sample(t)
        self.refresh_inf()
        return finish

    def refresh_inf(self):
        self.refresh_increment_inf("launch cost", self.cost)
        self.cost = 0

    @staticmethod
    def load(json, process_unit):
        if 'launch' not in json:
            return None
        else:
            json = json['launch']
        if json["type"] == "launch":
            return Launch(launch_time=Distribution.load(json, "launch_time"),
                          launch_cost=Distribution.load(json, "launch_cost"),
                          process_unit=process_unit)
        else:
            raise NotImplementedError


class Conversion(ProductionProcess):

    def __init__(self,
                 conversion_prepare_cost: Distribution,
                 conversion_rate_per_day: Distribution, conversion_cost_per_day: Distribution,
                 conversion_rate: Distribution, conversion_cost: Distribution,
                 conversion_period: Distribution,
                 produce_delay: Distribution,
                 process_unit):
        super().__init__(process_unit)
        self.conversion_prepare_cost = conversion_prepare_cost
        self.conversion_rate_per_day = conversion_rate_per_day
        self.conversion_cost_per_day = conversion_cost_per_day
        self.conversion_rate = conversion_rate
        self.conversion_cost = conversion_cost
        self.conversion_period = conversion_period
        self.produce_delay = produce_delay
        self.prod = None
        # self.obs_inf_k = 'conversion_' + str(self.process_unit.m_id)

        self.conversion_prepare_cost_v = 0
        self.conversion_finish_cost_v = 0
        self.conversion_daily_cost = 0
        self.obs_inf = {}

    def reset(self):
        super().reset()
        self.conversion_prepare_cost_v = 0
        self.conversion_finish_cost_v = 0
        self.conversion_daily_cost = 0
        self.obs_inf = {}
        self.prod = None

    def _justify(self, prod: Material, t):
        if t == 0:
            prepare_cost = self.get_distribution("conversion_prepare_cost", prod).sample(t)
            self.cost += prepare_cost
            self.conversion_prepare_cost_v += prepare_cost
            self.refresh_increment_inf("prepare cost", prepare_cost)
        time = self.get_distribution("conversion_period", prod).sample(t)
        produce_delay = self.get_distribution("produce_delay", prod).sample(t)
        output_rate = 0
        if t >= produce_delay:
            if self.prod is None:
                output_rate += self.get_distribution("conversion_rate_per_day", prod).sample(t)
            else:
                raise NotImplementedError
        daily_cost = self.get_distribution("conversion_cost_per_day", prod).sample(t)
        self.cost += daily_cost
        self.conversion_daily_cost += daily_cost
        self.refresh_increment_inf("daily cost", daily_cost)
        finish = t >= time
        if finish:
            finish_cost = self.get_distribution("conversion_cost", prod).sample(t)
            self.cost += finish_cost
            self.conversion_finish_cost_v += finish_cost
            self.refresh_increment_inf("finish cost", finish_cost)
            if self.prod is None:
                output_rate += self.get_distribution("conversion_rate", prod).sample(t)
            else:
                raise NotImplementedError
        self.prod = prod.conversion(self.process_unit.consume_material and finish, output_rate)
        self.refresh_increment_inf("total cost", self.cost)
        self.cost = 0
        return finish

    def refresh_inf(self):
        pass

    def convert(self, prod):
        return

    @staticmethod
    def load(json, process_unit):
        if 'conversion' not in json:
            return None
        else:
            json = json['conversion']
        if json["type"] == "conversion":
            # print(json)
            return Conversion(conversion_prepare_cost=Distribution.load(json, "conversion_prepare_cost"),
                              conversion_rate_per_day=Distribution.load(json, "conversion_rate_per_day"),
                              conversion_cost_per_day=Distribution.load(json, "conversion_cost_per_day"),
                              conversion_rate=Distribution.load(json, "conversion_rate"),
                              conversion_cost=Distribution.load(json, "conversion_cost"),
                              conversion_period=Distribution.load(json, "conversion_period"),
                              produce_delay=Distribution.load(json, "produce_delay"),
                              process_unit=process_unit)
        else:
            raise NotImplementedError


class Store(Process):

    def reset(self):
        super().reset()
        self.inventory = {}
        self.total = 0
        self.prod_total = {}
        self.wastage = 0
        self.wast_cost = 0
        self.inventory_cost = 0

    def refresh_inf(self):
        pass

    def __init__(self, quality_guarantee_period: Distribution,
                 inventory_hold_cost_per_unit_per_unit_date: Distribution,
                 wastage_rate_per_unit: Distribution, process_unit):
        super().__init__(process_unit)
        self.quality_guarantee_period = quality_guarantee_period
        self.inventory_hold_cost_per_unit_per_unit_date = inventory_hold_cost_per_unit_per_unit_date
        self.wastage_rate_per_unit = wastage_rate_per_unit
        # self.obs_inf_k = "store_" + str(self.process_unit.m_id)

        self.inventory = {}
        self.total = 0
        self.prod_total = {}
        self.wastage = 0
        self.wast_cost = 0
        self.inventory_cost = 0

    def store(self, product):
        prod = product.m_id.get_level_value(1)
        if prod not in self.inventory:
            self.inventory[prod] = []
        if prod not in self.prod_total:
            self.prod_total[prod] = 0
        self.inventory[prod].append(product)
        self.add_quantity(prod, product.quantity)

    def refresh_total_quantity(self):
        self.total = 0
        for prod in self.inventory:
            self.prod_total[prod] = 0
            for s in self.inventory[prod]:
                self.total += s.quantity
                self.prod_total[prod] += s.quantity

    def add_quantity(self, prod, quantity):
        assert quantity >= 0
        self.prod_total[prod] += quantity
        self.total += quantity
        if self.total <= 0 or self.prod_total[prod] <= 0:
            self.refresh_total_quantity()
        assert self.total >= 0 and self.prod_total[prod] >= 0, (prod, quantity, self.prod_total[prod])

    def minus_quantity(self, prod, quantity):
        assert quantity >= 0
        self.prod_total[prod] -= quantity
        self.total -= quantity
        if self.total <= 0 or self.prod_total[prod] <= 0:
            self.refresh_total_quantity()
        assert self.total >= 0 and self.prod_total[prod] >= 0, (prod, quantity, self.total, self.prod_total[prod])

    def fetch(self, prod, quantity):
        if prod not in self.prod_total or quantity <= 0 or len(self.inventory[prod]) == 0 or self.prod_total[prod] == 0:
            return 0
        else:
            ret = 0
            while len(self.inventory[prod]) > 0:
                buff = self.inventory[prod][0]
                if buff.quantity <= quantity:
                    delta = buff.quantity if buff.quantity < self.prod_total[prod] else self.prod_total[prod]
                    self.minus_quantity(prod, delta)
                    ret += delta
                    quantity -= delta
                    self.inventory[prod].remove(buff)
                else:
                    self.minus_quantity(prod, quantity)
                    buff.quantity -= quantity
                    ret += quantity
                    quantity = 0
                n = self.fetch(prod, quantity)
                if n == 0:
                    break
            return ret

    def remove_overdue_product(self, no_need_to_check: dict = None):
        if no_need_to_check is None:
            no_need_to_check = {}
            for prod, _ in self.inventory.items():
                no_need_to_check[prod] = False
                return self.remove_overdue_product(no_need_to_check)
        else:
            no_need = True
            for prod, _no_need in no_need_to_check.items():
                sub = Substance.substance[MID(SUBSTANCE_ID_PREFIX + "_{}_2".format(prod))]
                if not _no_need:
                    no_need_to_check[prod] = len(self.inventory[prod]) == 0 or ((self.process_unit.env.time -
                                                                                 self.inventory[prod][
                                                                                     0].born_time) <= self.get_distribution(
                        "quality_guarantee_period", sub).sample(self.process_unit.get_time()))
                _no_need = no_need_to_check[prod]
                if not _no_need:
                    # print("wast")
                    buff = self.inventory[prod][0]
                    wast_quantity = buff.quantity
                    self.wastage += wast_quantity
                    self.minus_quantity(prod, wast_quantity)
                    self.wast_cost += wast_quantity * self.get_distribution("wastage_rate_per_unit", sub).sample(
                        self.process_unit.get_time())
                    self.inventory[prod].remove(buff)
                no_need = no_need and _no_need
            if no_need:
                self.set_obs_inf(self.obs_inf_k, {
                    "wastage": self.wastage,
                    "wast cost": self.wast_cost
                })
                return True
            else:
                return self.remove_overdue_product(no_need_to_check)
        pass

    def refresh(self):
        self.remove_overdue_product()
        for prod, inventory in self.prod_total.items():
            sub = Substance.substance[MID(SUBSTANCE_ID_PREFIX + "_{}_2".format(prod))]
            self.inventory_cost += self.total * self.get_distribution("inventory_hold_cost_per_unit_per_unit_date",
                                                                      sub).sample(self.process_unit.get_time())
        obs = self.get_obs_inf(self.obs_inf_k, "inventory cost", {"inventory cost": self.inventory_cost})
        obs["inventory cost"] = self.inventory_cost
        # for prob, inventory in self.inventory.items():
        pass

    @staticmethod
    def load(json, process_unit):
        if json["type"] == "product":
            return Store(quality_guarantee_period=Distribution.load(json, "quality_guarantee_period"),
                         inventory_hold_cost_per_unit_per_unit_date=Distribution.load(json,
                                                                                      "inventory_hold_cost_per_unit_per_unit_date"),
                         wastage_rate_per_unit=Distribution.load(json, "wastage_rate_per_unit"),
                         process_unit=process_unit)
        else:
            raise NotImplementedError


class Sale(Process):

    def reset(self):
        super().reset()
        self.sales_volume = 0
        self.backlog_decay = 0
        self.backlog_penalty = 0

    def refresh_inf(self):
        pass

    def __init__(self, price_per_unit: Distribution,
                 backlog_penalty_cost_per_unit_per_day: Distribution,
                 backlog_decay_per_half_year: float, process_unit):
        super().__init__(process_unit)
        self.price_per_unit = price_per_unit
        self.backlog_penalty_cost_per_unit_per_day = backlog_penalty_cost_per_unit_per_day
        self.backlog_decay_per_half_year = 1 - math.pow(backlog_decay_per_half_year, 1 / 180)
        # self.obs_inf_k = "sale_" + str(self.process_unit.m_id)

        self.sales_volume = 0
        self.backlog_decay = 0
        self.backlog_penalty = 0

    def sale(self, product_storage_unit, customer_unit):
        total_demand = 0
        total_demand_fit = 0
        for prod, d in customer_unit.demand.demand_dict.items():
            ret = product_storage_unit.fetch(prod, d["not_fit"])
            d["fit"] += ret
            total_demand += d["all"]
            total_demand_fit += d["fit"]
            d["not_fit"] -= ret
            sub = Substance.substance[MID(SUBSTANCE_ID_PREFIX + "_{}_2".format(prod))]
            self.sales_volume += ret * self.get_distribution("price_per_unit", sub).sample(self.process_unit.get_time())
            self.backlog_penalty += d["not_fit"] * self.get_distribution("backlog_penalty_cost_per_unit_per_day",
                                                                         sub).sample(self.process_unit.get_time())
            decay = d["not_fit"] * self.backlog_decay_per_half_year
            self.backlog_decay += decay
            d["decay"] += decay
            d["not_fit"] -= decay
        # self.set_obs_inf(self.obs_inf_k, {"sales volume": self.sales_volume, "backlog decay": self.backlog_decay,
        #                                   "backlog penalty": self.backlog_penalty,
        #                                   "csl": total_demand_fit / total_demand * 100})

        product_storage_unit.store.refresh()
        store: Store = product_storage_unit.store
        demand: Demand = customer_unit.demand
        _all = {}
        fit = {}
        not_fit = {}
        decay = {}
        csl = {}
        inventory = {}
        for p, d in demand.demand_dict.items():
            _all[p] = d['all']
            fit[p] = d['fit']
            not_fit[p] = d['not_fit']
            decay[p] = d['decay']
            csl[p] = d['fit'] / d['all']
            inventory[p] = store.prod_total[p] if p in store.prod_total else 0
        self.set_obs_inf(self.obs_inf_k, {"sales volume": self.sales_volume, "backlog decay": self.backlog_decay,
                                          "backlog penalty": self.backlog_penalty,
                                          "csl": total_demand_fit / total_demand * 100,
                                          "all order values": _all, "fit order values": fit,
                                          "not fit order values": not_fit,"decay values": decay,
                                          "inventory values": inventory,
                                          "csl s": csl, "current time":self.process_unit.get_time()})

    @staticmethod
    def load(json, process_unit):
        if json["type"] == "sale":
            # print(json)
            return Sale(price_per_unit=Distribution.load(json, "price_per_unit"),
                        backlog_penalty_cost_per_unit_per_day=Distribution.load(json,
                                                                                "backlog_penalty_cost_per_unit_per_day"),
                        backlog_decay_per_half_year=(1 if "backlog_decay_per_half_year" not in json else float(
                            json["backlog_decay_per_half_year"])),
                        process_unit=process_unit)
        else:
            raise NotImplementedError


class Demand(Process):

    def reset(self):
        super().reset()
        self.demand_dict = {}  # 用于保存累计值
        for prod, dist in self.demand_json.items():
            self.demand_dict[prod] = {"all": 0, "fit": 0, "not_fit": 0, "decay": 0}

    def refresh_inf(self):
        pass

    def get_demand_u(self):
        u_table = []
        for prod, dist in self.demand_dist.items():
            if isinstance(dist, NormalDistribution):
                u_table.append(dist.mean.calculate(0))
            else:
                raise NotImplementedError
        return u_table

    def __init__(self, demand_json: dict, demand_scale, process_unit):
        super().__init__(process_unit)
        self.demand_json = demand_json
        self.demand_dist = {}  # 用于保存分布
        self.demand_scale = demand_scale

        self.demand_dict = {}  # 用于保存累计值
        for prod, dist in demand_json.items():
            self.demand_dist[prod] = Distribution.load(demand_json, prod)
            self.demand_dict[prod] = {"all": 0, "fit": 0, "not_fit": 0, "decay": 0}
        pass

    def generate(self):
        for prod, dist in self.demand_dist.items():
            dist: Distribution
            _demand = dist.sample(self.process_unit.get_time()) * self.demand_scale
            self.demand_dict[prod]["all"] += _demand
            self.demand_dict[prod]["not_fit"] += _demand

    def fit(self, prob, quantity):
        self.demand_dict[prob]["fit"] += quantity


class Scheduling(Process):

    def reset(self):
        super().reset()
        self.obs = []
        for i in range(self.prod_feature_dim * len(self.scheduling_valid_prod) + self.env_feature_dim):
            self.obs.append(0)
        return self.obs

    def __init__(self, scheduling_period: Distribution,obs_define: dict, reward_define, process_unit):
        super().__init__(process_unit)
        self.scheduling_period = scheduling_period
        # self.obs_inf_k = 'scheduling_' + str(self.process_unit.m_id)
        self.product_obs_define = obs_define["product"]
        self.prod_feature_dim = len(self.product_obs_define)
        self.env_obs_define = obs_define["env"]
        self.env_feature_dim = len(self.env_obs_define)
        self.reward_define = reward_define
        self.scheduling_valid_prod = []
        for sub_id in Substance.substance:
            if sub_id.get_level_value(1) not in self.scheduling_valid_prod:
                self.scheduling_valid_prod.append(sub_id.get_level_value(1))

        self.last_scheduling_time = 0
        self.last_profit = 0
        self.delta_time = 0
        self.delta_profit = 0
        self.last_action = None
        self.action = None
        self.obs = []
        for i in range(self.prod_feature_dim * len(self.scheduling_valid_prod) + self.env_feature_dim):
            self.obs.append(0)

    def on_new_scheduling(self, action):
        self.last_scheduling_time = self.process_unit.get_time()
        self.last_profit = self.process_unit.env.variables.get_variable("Profit")
        self.delta_time = 0
        self.delta_profit = 0
        self.obs = []
        self.last_action = action
        self.action = action

    def on_scheduling_finished(self):
        self.delta_time = self.process_unit.get_time() - self.last_scheduling_time
        self.delta_profit = self.process_unit.env.variables.get_variable("Profit") - self.last_profit

        last_action = self.last_action
        action = self.action
        delta_time = self.delta_time
        self.set_obs_inf(self.obs_inf_k, {"delta profit": self.delta_profit, "last_action":last_action,"action":action,
                                          "delta time":delta_time})
        # self.set_obs_inf(self.obs_inf_k, {"all order values": _all, "fit order values": fit,
        #                                   "not fit order values": not_fit,"decay values": decay,
        #                                   "inventory values": inventory,"delta profit": self.delta_profit,
        #                                   "csl s": csl, "last_action":last_action,"action":action,
        #                                   "delta time":delta_time, "current time":self.process_unit.get_time()})
        for p in self.scheduling_valid_prod:
            self.set_obs_inf(self.obs_inf_k, {"p": p}, replace=False)
            for product_obs_var in self.product_obs_define:
                self.obs.append(self.process_unit.env.variables.get_variable(product_obs_var))
        for env_obs_ver in self.env_obs_define:
            self.obs.append(self.process_unit.env.variables.get_variable(env_obs_ver))

    def scheduling(self, action):
        if isinstance(action, int):
            if self.process_unit.needs_scheduling():
                self.on_new_scheduling(action)
            else:
                action = None
            self.refresh_obs_inf_keys()
            self.process_unit.get_latter_unit().process(action)
            self.process_unit.env.time += 1
            self.set_obs_inf("scheduling", {"time": self.process_unit.env.time})
            done = self.process_unit.get_time() >= self.get_distribution("scheduling_period", None).sample()
            if self.process_unit.needs_scheduling() or done:
                self.on_scheduling_finished()
                return self.obs, self.process_unit.env.variables.get_variable(self.reward_define), done
            return self.obs, 0, done
        else:
            raise NotImplementedError

    @staticmethod
    def load(json, process_unit):
        if "type" not in json:
            if 'scheduling' not in json:
                return None
            else:
                json = json['scheduling']
        if json["type"] == "scheduling":
            # print(json)
            p = Scheduling(scheduling_period=Distribution.load(json, "scheduling_period"),
                           obs_define=json["obs"],
                           reward_define=json["reward"],
                           process_unit=process_unit)
            return p
        else:
            raise NotImplementedError



import time
import datetime

top_task = None


class TaskTime(object):

    def __init__(self, name, as_top_task=False):
        global top_task
        self.name = name
        self.time_span_list = []
        self.start_time = -1
        self.end_time = -1
        self.total = 0
        self.already = 0
        self.log = True
        if as_top_task or top_task is None:
            top_task = self

    def start(self):
        assert (self.start_time == -1), self.name
        self.start_time = time.time()

    def end(self):
        assert (self.end_time == -1), self.name
        self.end_time = time.time()
        self.task_exec_once()

    def task_exec_once(self):
        global top_task
        delta = self.end_time - self.start_time
        self.time_span_list.append([self.start_time, self.end_time, delta])
        self.total += delta
        self.end_time = self.start_time = -1
        tcm = self.total / len(self.time_span_list)
        top_task.refresh()
        if self.log:
            print('\t\t\t\t\t task:%s executed %d times,tct:%e s,tcm:%.e s, tpt:%.4f%%,tpm:%.4f%%,top:%s' %
                  (self.name, len(self.time_span_list), self.total, tcm, self.total * 100. / top_task.already,
                   tcm * 100. / top_task.already, top_task.name))

    def refresh(self):
        self.already = time.time() - self.start_time
        if self.already == 0:
            print("warning: already equals zero")
            self.already = 1


class TimeCalculator(object):

    def __init__(self):
        self.task_time_dict = {}
        self.log = True
        pass

    def st(self, name, as_top_task=False):
        if name not in self.task_time_dict:
            self.task_time_dict[name] = TaskTime(name, as_top_task)
            self.task_time_dict[name].log = self.log
        tt: TaskTime = self.task_time_dict[name]
        tt.start()

    def ed(self, name):
        assert name in self.task_time_dict
        tt: TaskTime = self.task_time_dict[name]
        tt.end()

    def set_log(self, log=True):
        self.log = log
        for n, t in self.task_time_dict.items():
            t: TaskTime
            t.log = log


time_calculator = TimeCalculator()
if __name__ == '__main__':
    time_calculator.st('loop')
    for j in range(1111):
        time_calculator.st('loop1')
        for i in range(100000):
            # print(i)
            pass
        time_calculator.ed('loop1')
        time_calculator.st('loop2')
        for i in range(100000):
            # print(i)
            pass
        time_calculator.ed('loop2')
    time_calculator.ed('loop')



import copy

from elsp_env_manager.base import MID, Distribution
from elsp_env_manager.base.constants import SUBSTANCE_ID_PREFIX
from elsp_env_manager.base.model_manager import load_model


class Substance(object):
    substance: dict = {}

    def __init__(self, _id: MID, model_json: dict = None):
        self.m_id = _id
        self.incomplete_operation = [] if model_json is None or "incomplete_operation" not in model_json else model_json["incomplete_operation"]
        self.quantity = 0
        self.model_json = model_json

    def same_series(self, l1_or_id):
        return self.m_id.get_level_value(1) == l1_or_id if not isinstance(l1_or_id, MID) else self.m_id.get_level_value(
            1) == l1_or_id.get_level_value(1)

    def next_production_substance(self):
        return Substance.instance(self.m_id.next_production_id(), 0)

    @staticmethod
    def load(json_or_json_path):
        if isinstance(json_or_json_path, dict):
            json = json_or_json_path
        else:
            json = load_model(json_or_json_path)
        for p_id, p in json.items():
            if p is None:
                m_id = MID(SUBSTANCE_ID_PREFIX + "_{}".format(p_id))
                Substance.substance[m_id] = Substance(m_id)
                continue
            for sp_id, sp in p.items():
                m_id = MID(SUBSTANCE_ID_PREFIX + "_{}".format(p_id) + "_{}".format(sp_id))
                if sp["type"] == "material":
                    Substance.substance[m_id] = Material(_id=m_id, model_json=sp)
                elif sp["type"] == 'product':
                    Substance.substance[m_id] = Product(_id=m_id, model_json=sp)
                else:
                    raise NotImplementedError
                # Product(m_id, [], quantity=0)

            # products[m_id] = Product(m_id, [], quantity=0)

    @staticmethod
    def instance(_id: MID, quantity: float):
        if _id not in Substance.substance and _id.id_level >= 2:
            _id = _id.upper()
        else:
            assert NotImplementedError
        if _id not in Substance.substance:
            return None
        sub: Substance = copy.deepcopy(Substance.substance[_id])
        sub.quantity = quantity
        return sub

    @staticmethod
    def get_instance_by_l1(l1: str):
        return Substance.instance(MID("Substance_{}".format(l1)), 0)

    @staticmethod
    def get_instance(id_or_l1):
        if id_or_l1 not in Substance.substance:
            sub = Substance.get_instance_by_l1(id_or_l1)
        else:
            sub = Substance.substance[id_or_l1]
        return sub


class Material(Substance):

    def __init__(self, _id: MID, model_json: dict = None):
        super().__init__(_id, model_json)

    def conversion(self, consume_material, rate):
        res: Substance = self.next_production_substance()
        res.quantity = rate * self.quantity
        if consume_material:
            self.quantity = 0
        return res


class Product(Substance):

    def __init__(self, _id, model_json: dict = None):
        super().__init__(_id, model_json)
        self.born_time = 0



import json
import os


def load_model(model_name: str, env_path=None):
    if env_path is not None:
        if not model_name.endswith(".json"):
            model_name = model_name + ".json"
        assert model_name in os.listdir(env_path)
        model_file = open(os.path.join(env_path, model_name))
    else:
        model_file = open(model_name)
    model = json.loads(model_file.read())
    model_file.close()
    return model




CUSTOMER_UNITS_FILE_NAME = "customer_units.json"
PROCESS_LINES_FILE_NAME = "process_lines.json"
PRODUCTION_UNITS_FILE_NAME = "production_units.json"
PRODUCTS_FILE_NAME = "products.json"
SALE_UNITS_FILE_NAME = "sale_units.json"
SCHEDULING_UNITS_FILE_NAME = "scheduling_units.json"
STORAGE_UNITS_FILE_NAME = "storage_units.json"
OBS_DEFINE_FILE_NAME = "obs_define.json"
VARIABLE_MAP_FILE_NAME = "variable_map.json"
VARIABLE_DEFINE_FILE_NAME = "variable_define.json"

CUSTOMER_UNITS_ID_PREFIX = "CustomerUnit"
PROCESS_LINES_ID_PREFIX = "ProcessLine"
PRODUCTION_UNITS_ID_PREFIX = "ProductionUnit"
SUBSTANCE_ID_PREFIX = "Substance"
SALE_UNITS_ID_PREFIX = "SaleUnit"
SCHEDULING_UNITS_ID_PREFIX = "SchedulingUnit"
STORAGE_UNITS_ID_PREFIX = "StorageUnit"
PARALLEL_COLLECTION_UNITS_ID_PREFIX = "ParallelCollectionUnit"

















from abc import ABC, abstractmethod
from elsp_env_manager.base import MID, Substance, Material, Product
from elsp_env_manager.base.constants import *
from elsp_env_manager.base.model_manager import load_model
from elsp_env_manager.process import Setup, Launch, Conversion, Demand, Sale, Store
from elsp_env_manager.process.process import Scheduling, Failure


class ProcessUnit(ABC):

    def __init__(self, m_id: MID, model_json, env):
        self._former_unit = None
        self._latter_unit = None
        self.env = env
        self.process_start_time = env.time
        self.last_process_start_time = env.time
        self.m_id = m_id
        self.model_json = model_json

    def get_time(self):
        return self.env.time - self.process_start_time

    def set_latter_unit(self, unit):
        self._latter_unit = unit
        unit._former_unit = self

    def set_former_unit(self, unit):
        self._former_unit = unit
        unit._latter_unit = self

    def get_latter_unit(self):
        return self._latter_unit

    def get_former_unit(self):
        return self._former_unit

    @abstractmethod
    def process(self, *args):
        pass

    @abstractmethod
    def reset(self):
        pass


class ProductionUnit(ProcessUnit):
    product_units = {}

    def reset(self):
        if self.setup is not None:
            self.setup.reset()
        if self.launch is not None:
            self.launch.reset()
        if self.conversion is not None:
            self.conversion.reset()
        if self.failure is not None:
            self.failure.reset()
        self.idle = True

    def __init__(self, m_id: MID, model_json, env):
        super().__init__(m_id, model_json, env)
        self.setup = Setup.load(model_json, self)
        self.launch = Launch.load(model_json, self)
        self.conversion = Conversion.load(model_json, self)
        self.failure = Failure.load(model_json, self)
        self.process_list = ['setup','launch','conversion','failure']
        self.available_material_list = []
        self.consume_material = model_json["consume_material"]
        self.idle = True

    def __setattr__(self, key, value):
        self.__dict__[key] = value
        if key == 'idle':
            self.conversion.set_obs_inf(self.conversion.obs_inf_k, {'idle': value}, False)

    def process(self, material: Substance):
        if material.model_json is None:
            self.idle = True
            if self.get_latter_unit() is not None:
                self.get_latter_unit().process(material)
            else:
                raise RuntimeError
            return True
        else:
            assert self.m_id.get_level_value(1) in material.model_json["incomplete_operation"]
            if self.idle:
                self.last_process_start_time = self.process_start_time
                self.process_start_time = self.env.time
                self.idle = False
            # print(material.model_json)
            if self.setup is not None:
                self.setup.process(material)
            if self.launch is not None:
                self.launch.process(material, self.setup)
            if self.failure is not None:
                failed = self.failure.process(material, self.launch)
                if failed:
                    self.reset()
                    self.get_latter_unit().reset()
                    return failed

            finished = self.conversion.process(material, self.launch) if self.conversion is not None else True
            if isinstance(self.conversion.prod, Product):
                self.conversion.prod.born_time = self.env.time
            self.get_latter_unit().process(self.conversion.prod)

            self.conversion.prod = None
            if finished:
                if self.setup is not None:
                    self.setup.reset()
                if self.launch is not None:
                    self.launch.reset()
                if self.conversion is not None:
                    self.conversion.reset()
                self.idle = True
            return finished

    @staticmethod
    def load(json_or_json_path, env):
        if isinstance(json_or_json_path, dict):
            json = json_or_json_path
        else:
            json = load_model(json_or_json_path)
        for u_id, u in json.items():
            m_id = MID(_id="ProductionUnit_" + u_id)
            if u["type"] == 'continuous':
                # print(u_id, u)
                ProductionUnit.product_units[m_id] = ContinuousProductionUnit(m_id=m_id, model_json=u, env=env)
            elif u["type"] == 'batch':
                ProductionUnit.product_units[m_id] = BatchProductionUnit(m_id=m_id, model_json=u, env=env)
            else:
                raise NotImplementedError

    @staticmethod
    def get_unit_by_id(_id):
        u = ProductionUnit.product_units[MID(str(_id))] if not isinstance(_id, MID) else \
            ProductionUnit.product_units[_id]
        return u


class ContinuousProductionUnit(ProductionUnit):

    def reset(self):
        super().reset()
        self.converting_material = None

    def __init__(self, m_id: MID, model_json, env):
        super().__init__(m_id, model_json, env)
        self.converting_material = None

    def process(self, material):
        if self.converting_material is not None and material is not None:
            raise RuntimeError
        if self.converting_material is None:
            self.converting_material = material
        res = super().process(self.converting_material)
        if res:
            self.converting_material = None
        return res


class BatchProductionUnit(ProductionUnit):

    def reset(self):
        for c_p_u in self.c_p_u_list:
            c_p_u.reset()
        for m in list(self.c_p_u_dict.keys()):
            self.c_p_u_dict.pop(m)

    def __init__(self, m_id: MID, model_json, env):
        super().__init__(m_id, model_json, env)
        self.c_p_u_dict = {}
        self.c_p_u_list = []

    def get_c_p_u(self):
        for c_p_u in self.c_p_u_list:
            if c_p_u.idle:
                return c_p_u
        c_p_u = ContinuousProductionUnit(self.m_id, self.model_json, self.env)
        self.c_p_u_list.append(c_p_u)
        return c_p_u

    def process(self, material: Material):
        m_to_remove = []
        for m, c_p_u in self.c_p_u_dict.items():
            c_p_u: ContinuousProductionUnit
            finished = c_p_u.process(m if c_p_u.idle else None)
            if finished:
                # print(material, m, m.quantity)
                m_to_remove.append(m)
        for m in m_to_remove:
            self.c_p_u_dict.pop(m)
        if material is not None and material.model_json is not None and material.quantity == 1:
            c_p_u = self.get_c_p_u()
            c_p_u.set_latter_unit(self.get_latter_unit())
            finished = c_p_u.process(material)
            if not finished:
                self.c_p_u_dict[material] = c_p_u
        return len(self.c_p_u_dict) == 0


class StorageUnit(ProcessUnit):
    storage_units = {}

    def process(self, material: Substance):
        if material.model_json is None:
            print("nothing to store")
            if self.get_latter_unit() is not None:
                self.get_latter_unit().process(material)
            else:
                print("warning: unit without follow unit")
        else:
            if material.quantity == 0:
                self.get_latter_unit().process(material)
            else:
                raise NotImplementedError
        pass

    @staticmethod
    def load(json_or_json_path, env):
        if isinstance(json_or_json_path, dict):
            json = json_or_json_path
        else:
            json = load_model(json_or_json_path)
        for u_id, u in json.items():
            m_id = MID(_id="StorageUnit_" + u_id)
            # print(u)
            if u["type"] == "material":
                StorageUnit.storage_units[m_id] = MaterialStorageUnit(m_id, u, env)
            elif u["type"] == "product":
                StorageUnit.storage_units[m_id] = ProductStorageUnit(m_id, u, env)
            else:
                raise NotImplementedError

    @staticmethod
    def get_unit_by_id(_id):
        return StorageUnit.storage_units[MID(str(_id))] if not isinstance(_id, MID) else \
            StorageUnit.storage_units[_id]

    @abstractmethod
    def fetch(self, _id: MID, quantity: float) -> Substance:
        pass


class MaterialStorageUnit(StorageUnit):

    def reset(self):
        pass

    def fetch(self, _id: MID, quantity: float) -> Substance:
        if self.infinite:
            return Substance.instance(_id, quantity=quantity)
        else:
            raise NotImplementedError

    def __init__(self, m_id: MID, model_json, env):
        super().__init__(m_id, model_json, env)
        self.infinite = model_json["infinite"]

    def process(self, material):
        if material is None:
            self.get_latter_unit().process(None)
        else:
            self.get_latter_unit().process(self.fetch(MID("Substance_{}_{}".format(material, 0)), 1))


class ProductStorageUnit(StorageUnit):

    def reset(self):
        self.store.reset()

    def __init__(self, m_id: MID, model_json, env):
        super().__init__(m_id, model_json, env)
        self.store = Store.load(model_json, self)

    def process(self, product: Product):
        if product.quantity == 0:
            # self.get_latter_unit().process(product)
            pass
        else:
            self.store.store(product=product)
            pass

    def fetch(self, prod, quantity):
        return self.store.fetch(prod, quantity)


class SaleUnit(ProcessUnit):
    sale_units = {}

    def reset(self):
        self.sale.reset()

    def get_latter_unit(self):
        u = super().get_latter_unit()
        return u

    def __init__(self, m_id: MID, model_json, env):
        super().__init__(m_id, model_json, env)
        self.sale = Sale.load(model_json, self)
        pass

    def process(self, product: Substance):
        if hasattr(product, "model_json") and product.model_json is None:
            print("nothing to sale")
            if self.get_latter_unit() is not None:
                self.get_latter_unit().process(product)
            else:
                raise NotImplementedError
        elif isinstance(product, tuple):
            customer_unit: CustomerUnit = product[1][1]
            product_storage_unit = product[1][0].last_unit()
            self.sale.sale(customer_unit=customer_unit, product_storage_unit=product_storage_unit)
            pass
        else:
            raise NotImplementedError
        pass

    @staticmethod
    def load(json_or_json_path, env):
        if isinstance(json_or_json_path, dict):
            json = json_or_json_path
        else:
            json = load_model(json_or_json_path)
        for u_id, u in json.items():
            m_id = MID(_id="SaleUnit_" + u_id)
            SaleUnit.sale_units[m_id] = SaleUnit(m_id, u, env)

    @staticmethod
    def get_unit_by_id(_id):
        return SaleUnit.sale_units[MID(str(_id))] if not isinstance(_id, MID) else \
            SaleUnit.sale_units[_id]


class CustomerUnit(ProcessUnit):
    customer_units = {}

    def reset(self):
        self.demand.reset()

    def __init__(self, m_id: MID, model_json, demand_scale, env):
        super().__init__(m_id, model_json, env)
        self.demand: Demand = Demand(model_json["demand"], demand_scale, self)

    def process(self, *args):
        self.demand.generate()
        self.get_latter_unit().process(*args)
        pass

    @staticmethod
    def load(json_or_json_path, demand_scale, env):
        if isinstance(json_or_json_path, dict):
            json = json_or_json_path
        else:
            json = load_model(json_or_json_path)
        for u_id, u in json.items():
            m_id = MID(_id=CUSTOMER_UNITS_ID_PREFIX + "_" + u_id)
            CustomerUnit.customer_units[m_id] = CustomerUnit(m_id, u, demand_scale, env)


class SchedulingUnit(ProcessUnit):
    scheduling_units = {}

    def reset(self):
        self.obs = self.scheduling.reset()
        latter_unit = self.get_latter_unit()
        while True:
            latter_unit.reset()
            latter_unit = latter_unit.get_latter_unit()
            if isinstance(latter_unit, SchedulingUnit):
                break
            pass

    def __init__(self, m_id: MID, model_json, env):
        super().__init__(m_id, model_json, env)
        if model_json["production_unit"]["type"] == "production":
            self.production_unit = ProductionUnit.product_units[MID(PRODUCTION_UNITS_ID_PREFIX+"_{}".format(model_json["production_unit"]["id"]))]
            self.storage_unit = StorageUnit.storage_units[MID(STORAGE_UNITS_ID_PREFIX+"_1")]
            self.customer_unit = CustomerUnit.customer_units[MID(CUSTOMER_UNITS_ID_PREFIX+"_0")]
        else:
            raise NotImplementedError
        self.scheduling = Scheduling.load(model_json, self)
        
        self.obs = []
        self.reward = 0
        self.done = False
        self.inf = {}
        self.step_inf = {}
        self.action = None
        self.last_action = None

    def process(self, action):
        self.obs, self.reward, self.done = self.scheduling.scheduling(action)

    @staticmethod
    def is_idle_action(action):
        sub = Substance.get_instance_by_l1(action)
        return sub is not None and sub.model_json is None

    def needs_scheduling(self):
        return self.production_unit.idle

    @staticmethod
    def load(json_or_json_path, env):
        if isinstance(json_or_json_path, dict):
            json = json_or_json_path
        else:
            json = load_model(json_or_json_path)
        for u_id, u in json.items():
            m_id = MID(_id=SCHEDULING_UNITS_ID_PREFIX + "_" + u_id)
            SchedulingUnit.scheduling_units[m_id] = SchedulingUnit(m_id, u, env)



import random
import numpy as np

seed = None
random.seed(seed)
np.random.seed(seed)


class BSP2(object):
    def __init__(self, u_table):
        self.u_table = u_table
        self.prod_num = len(self.u_table)
        self.m = 0
        self.I_table = []
        for i in range(self.prod_num):
            self.I_table.append(0)
        self.Y1_table = []
        self.Y2_table = []
        self.Y3_table = []
        self.Y4_table = []
        for i in range(self.prod_num):
            self.Y1_table.append(0)
            self.Y2_table.append(0)
            self.Y3_table.append(0)
            self.Y4_table.append(0)

    def choose_action(self, y, last_m, inv):
        self.set_y(y)
        self.set_state(last_m, inv)
        return self.choose_product_to_produce()

    def choose_product_to_produce(self):
        m = self.m - 1
        if self.m > 0 and self.I_table[m] < self.Y4_table[m]:
            return self.m
        elif self.get_z1() is not []:
            z = self.get_z1()
            _min, min_index = -1, -1
            for i in z:
                run_out_time = self.I_table[i] / self.u_table[i]
                if _min == -1 or _min > run_out_time:
                    _min = run_out_time
                    min_index = i
                elif _min == run_out_time:
                    if random.random() >= 0.5:
                        min_index = i
            return min_index + 1
        elif self.m > 0 and self.I_table[m] < self.Y2_table[m]:
            return self.m
        elif self.get_z3() is not []:
            z = self.get_z3()
            _min, min_index = -1, -1
            for i in z:
                run_out_time = self.I_table[i] / self.u_table[i]
                if _min == -1 or _min > run_out_time:
                    _min = run_out_time
                    min_index = i
                elif _min == run_out_time:
                    if random.random() >= 0.5:
                        min_index = i
            return min_index + 1
        else:
            return 0

    def get_z1(self):
        z = []
        for i in range(self.prod_num):
            if self.I_table[i] <= self.Y1_table[i]:
                z.append(i)
        return z

    def get_z3(self):
        z = []
        for i in range(self.prod_num):
            if self.I_table[i] <= self.Y3_table[i]:
                z.append(i)
        return z

    def set_state(self, m, I_table):
        self.m = m
        self.I_table = I_table

    def set_policy_parameters(self, y1, y2, y3, y4):
        self.Y1_table = y1
        self.Y2_table = y2
        self.Y3_table = y3
        self.Y4_table = y4

    def set_policy_parameters_by_tables(self,tables):
        self.Y1_table = tables[0]
        self.Y2_table = tables[1]
        self.Y3_table = tables[2]
        self.Y4_table = tables[3]

    def get_y_tables(self):
        y = [self.Y1_table, self.Y2_table, self.Y3_table,
             self.Y4_table]
        return y

    def get_y(self):
        y = []
        for i in range(self.prod_num):
            y.append(self.Y1_table[i])
            y.append(self.Y2_table[i])
            y.append(self.Y3_table[i])
            y.append(self.Y4_table[i])
        return y

    def set_y(self, y):
        assert len(y) == self.prod_num * 4
        for i in range(self.prod_num):
            self.Y1_table[i] = y[i * 4 + 0]
            self.Y2_table[i] = y[i * 4 + 1]
            self.Y3_table[i] = y[i * 4 + 2]
            self.Y4_table[i] = y[i * 4 + 3]


if __name__ == "__main__":
    from elsp_env_manager.elsp_env import ELSPEnv
    from elsp_env_manager.eval_env import EvalEnv
    env = ELSPEnv("../../../elsp_env_00014", "../../../")
    agent = BSP2(env.get_demand_u())

    def policy_maker(last_action, obs):
        y = [1.5522879261492974, 61.5522879261493, 1.5522879261492974, 61.5522879261493, 0.0, 120.0, 0.0, 60.0, 0.0,
             5.759476184616176, 0.0, 5.759476184616176, 0.0, 60.0, 0.0, 0.0, 25.158372645803503, 25.158372645803503,
             25.158372645803503, 25.158372645803503, 0.0, 0.0, 0.0, 0.0]
        inv = []
        for i in range(agent.prod_num):
            inv.append(obs[3 * (i + 1) + 2])
        return agent.choose_action(y, last_m=last_action, inv=inv)

    def eval_data_getter(last_inf):
        profit = last_inf["Profit"]
        return profit

    eval_env = EvalEnv(policy_maker=policy_maker,
                       envs=[ELSPEnv("../../../elsp_env_00014", "../../../") for _ in range(20)], max_process=20,
                       eval_data_getter=eval_data_getter)
    res = eval_env.eval(100)
    mean = sum(res) / len(res)
    print(mean)
    eval_env.destroy()





import random


import gym
import numpy as np
import os
from elsp_env_manager.base import Product, MID, Variables
from elsp_env_manager.process_units.process_line import ProcessLine
from elsp_env_manager.process_units.process_units import ProductionUnit, SchedulingUnit, StorageUnit, SaleUnit, \
    CustomerUnit
from elsp_env_manager.base.constants import *
from utils_minan.TimeCalculator import time_calculator
try:
    import torch as th
    has_torch = True
except ImportError:
    has_torch = False

# seed = 7
# random.seed(seed)
# np.random.seed(seed)


class ELSPEnv(gym.Env):

    def __init__(self, env_path, cfg_path, demand_scale=1):
        self.env_path = env_path
        self.cfg_path = cfg_path
        self.time_unit = 'day'
        self.time = 0
        self.demand_scale = demand_scale
        Product.load(os.path.join(env_path, PRODUCTS_FILE_NAME))
        ProductionUnit.load(os.path.join(env_path, PRODUCTION_UNITS_FILE_NAME), self)
        StorageUnit.load(os.path.join(env_path, STORAGE_UNITS_FILE_NAME), self)
        CustomerUnit.load(os.path.join(env_path, CUSTOMER_UNITS_FILE_NAME), env=self, demand_scale=demand_scale)
        SaleUnit.load(os.path.join(env_path, SALE_UNITS_FILE_NAME), self)
        SchedulingUnit.load(os.path.join(env_path, SCHEDULING_UNITS_FILE_NAME), self)
        ProcessLine.load(os.path.join(env_path, PROCESS_LINES_FILE_NAME), self)
        assert len(SchedulingUnit.scheduling_units) == 1
        self.scheduling_unit: SchedulingUnit = SchedulingUnit.scheduling_units[MID(SCHEDULING_UNITS_ID_PREFIX + "_0")]
        self.variables = Variables.load(map_json_file=os.path.join(cfg_path, VARIABLE_MAP_FILE_NAME),
                                        define_json_file=os.path.join(cfg_path, VARIABLE_DEFINE_FILE_NAME),
                                        inf=self.scheduling_unit.inf)
        num = self.scheduling_unit.scheduling.obs
        self.observation_space = gym.spaces.Box(np.zeros_like(self.scheduling_unit.scheduling.obs),
                                                np.ones_like(self.scheduling_unit.scheduling.obs))
        self.action_space = gym.spaces.Discrete(len(self.scheduling_unit.scheduling.scheduling_valid_prod))
        setattr(self.action_space, 'flexible', True)
        self.recorded_inf = {}
        self.count_recorded_inf = {}
        self.count_started = {}

    def record(self, k, v):
        if k not in self.recorded_inf:
            self.recorded_inf[k] = []
        self.recorded_inf[k].append(v)

    def count_record(self, k):
        if k not in self.count_started or not self.count_started[k] or len(self.count_recorded_inf[k]) == 0:
            self.count_start(k)
            self.count_started[k] = True
        self.count_recorded_inf[k][-1] += 1

    def count_start(self, k):
        if k not in self.count_recorded_inf:
            self.count_recorded_inf[k] = []
        self.count_recorded_inf[k].append(0)

    def get_recorded_inf(self):
        buff = self.recorded_inf
        self.recorded_inf = {}
        ep_num = 0
        for k, _ in buff.items():
            ep_num = max(ep_num, len(_))
        for k, v in self.count_recorded_inf.items():
            buff[k] = v[0:ep_num]
            self.count_recorded_inf[k] = v[ep_num:]
        return buff

    def step(self, action):
        inf = self.scheduling_unit.inf
        self.count_record(str(action))
        while True:
            self.scheduling_unit.process(action)
            if self.scheduling_unit.needs_scheduling() or self.scheduling_unit.done:
                if self.scheduling_unit.done:
                    self.scheduling_unit.inf = {}
                    for k, v in self.count_started.items():
                        self.count_started[k] = False
                    self.record("Profit", self.variables.get_variable("Profit"))
                    inf["Profit"] = self.variables.get_variable("Profit")
                break
        return self.scheduling_unit.obs, self.scheduling_unit.reward, self.scheduling_unit.done, inf

    def reset(self):
        self.time = 0
        env_path = self.env_path
        cfg_path = self.cfg_path
        Product.load(os.path.join(env_path, PRODUCTS_FILE_NAME))
        ProductionUnit.load(os.path.join(env_path, PRODUCTION_UNITS_FILE_NAME), self)
        StorageUnit.load(os.path.join(env_path, STORAGE_UNITS_FILE_NAME), self)
        CustomerUnit.load(os.path.join(env_path, CUSTOMER_UNITS_FILE_NAME), self.demand_scale, self)
        SaleUnit.load(os.path.join(env_path, SALE_UNITS_FILE_NAME), self)
        SchedulingUnit.load(os.path.join(env_path, SCHEDULING_UNITS_FILE_NAME), self)
        ProcessLine.load(os.path.join(env_path, PROCESS_LINES_FILE_NAME), self)
        assert len(SchedulingUnit.scheduling_units) == 1
        self.scheduling_unit: SchedulingUnit = SchedulingUnit.scheduling_units[MID(SCHEDULING_UNITS_ID_PREFIX + "_0")]
        self.variables = Variables.load(map_json_file=os.path.join(cfg_path, VARIABLE_MAP_FILE_NAME),
                                        define_json_file=os.path.join(cfg_path, VARIABLE_DEFINE_FILE_NAME),
                                        inf=self.scheduling_unit.inf)
        self.scheduling_unit.reset()
        return self.scheduling_unit.obs

    def render(self, mode='human'):
        pass

    def get_demand_u(self):
        if len(CustomerUnit.customer_units) == 1:
            for k, v in CustomerUnit.customer_units.items():
                v: CustomerUnit
                return v.demand.get_demand_u()
        else:
            raise NotImplementedError

    def decode_features(self, features):
        # features = np.array(features)
        feature_dim = features.shape[-1]
        env_feature_dim = self.scheduling_unit.scheduling.env_feature_dim
        prod_feature_dim = self.scheduling_unit.scheduling.prod_feature_dim
        prod_num = (feature_dim - env_feature_dim) // prod_feature_dim
        if has_torch:
            _obs = th.split(features, [feature_dim - env_feature_dim, env_feature_dim], dim=1)
        else:
            _obs = np.split(features, [feature_dim - env_feature_dim], axis=1)
        s1 = _obs[0]
        s1 = s1.reshape((prod_num, env_feature_dim) if len(s1.shape) == 1 else (s1.shape[0], prod_num, prod_feature_dim))
        s2 = _obs[1]
        return s1, s2


if __name__ == "__main__":
    time_calculator.st("main")
    env = ELSPEnv(env_path="../elsp_env_00001", cfg_path="..", demand_scale=1)


    def log_inf(_inf):
        for k, v in _inf.items():
            if not isinstance(v, dict):
                print(k, v)
            else:
                for sk, sv in v.items():
                    print(k, sk, sv)


    for i in range(1):
        time_calculator.st("ep")
        obs = env.reset()
        print(obs)
        action = 1
        while True:
            obs, reward, done, inf = env.step(1 if False else action % 4)
            # print(action % 4, obs)
            # 36.04320240231678
            obs = np.array([obs, obs.copy()])
            s1, s2 = env.decode_features(obs)
            print(s1)
            action += 1
            # log_inf(inf)
            if done:
                break
        time_calculator.ed("ep")
        """print("Profit", env.variables.get_variable("Profit"))
        print("Revenue", env.variables.get_variable("Revenue"))
        print("Total costs", env.variables.get_variable("Total costs"))
        print("\tSeed", env.variables.get_variable("Seed"))
        print("\tUSP", env.variables.get_variable("USP"))
        print("\t\tReplacement ATF filters", env.variables.get_variable("Replacement ATF filters"))
        print("\t\tCell culture setup", env.variables.get_variable("Cell culture setup"))
        print("\tDSP", env.variables.get_variable("DSP"))
        print("\tChangeover", env.variables.get_variable("Changeover"))
        print("\tStorage", env.variables.get_variable("Storage"))
        print("\tBacklog penalties", env.variables.get_variable("Backlog penalties"))
        print("\tWastage", env.variables.get_variable("Wastage"))
        print("CSL", env.variables.get_variable("CSL"))"""
    """
        print(
            inf["sale_SaleUnit_0"]["sales_volume"] - inf["setup_ProductionUnit_0"]["setup cost"]
            - inf["sale_SaleUnit_0"]["backlog_penalty"] - inf["store_StorageUnit_1"]["inventory cost"] \
            - inf["launch_ProductionUnit_0"]["launch cost"] \
            - inf["conversion_ProductionUnit_0"]["total cost"] \
            - inf["conversion_ProductionUnit_1"]["total cost"]
        )"""
    print(env.get_recorded_inf())
    time_calculator.ed("main")



from elsp_env_manager.base.distribution import (
    UnaryPolynomial,
    Distribution, Deterministic, UniformDistribution, NormalDistribution
)
from elsp_env_manager.base.m_id import MID
from elsp_env_manager.base.substance import (
    Substance, Material, Product
)
from elsp_env_manager.base.variables import Variables




from abc import ABC, abstractmethod
import math
import numpy as np


class UnaryPolynomial(object):

    def __init__(self, polynomial_coefficient):
        assert not isinstance(polynomial_coefficient, dict), polynomial_coefficient
        # 兼容直接给定常数项
        if not isinstance(polynomial_coefficient, list):
            polynomial_coefficient = [polynomial_coefficient]
        self.polynomial_coefficient = polynomial_coefficient

    def calculate(self, x: float = 1):
        res = 0
        for i in range(len(self.polynomial_coefficient)):
            coe = self.polynomial_coefficient[i]
            if coe == 0:
                continue
            res += coe * (math.pow(x, i))
        return res


class Distribution(ABC):

    @abstractmethod
    def sample(self, x: float):
        pass

    @staticmethod
    def load(json, name):
        if name not in json:
            return Distribution.load({"0": 0}, "0")
        elif json[name] == "up_to_product":
            return json[name]
        json = json[name]
        if isinstance(json, list) and isinstance(json[0], str):
            if json[0] == 'n':
                return NormalDistribution(json[1], json[2])
            else:
                assert NotImplementedError
        else:
            return Deterministic(json)


class UniformDistribution(Distribution):

    def __init__(self, a_polynomial_coefficient: list, b_polynomial_coefficient: list):
        self.a: UnaryPolynomial = UnaryPolynomial(a_polynomial_coefficient)
        self.b: UnaryPolynomial = UnaryPolynomial(b_polynomial_coefficient)

    def sample(self, x: float = 1):
        return np.random.uniform(self.a.calculate(x), self.b.calculate(x))


class NormalDistribution(Distribution):

    def __init__(self, mean_polynomial_coefficient: list, std_polynomial_coefficient: list):
        self.mean: UnaryPolynomial = UnaryPolynomial(mean_polynomial_coefficient)
        self.std: UnaryPolynomial = UnaryPolynomial(std_polynomial_coefficient)

    def sample(self, x: float = 1):
        return np.random.normal(self.mean.calculate(x), self.std.calculate(x))


class Deterministic(Distribution):

    def __init__(self, base_coefficient: list):
        self.base: UnaryPolynomial = UnaryPolynomial(base_coefficient)

    def sample(self, x: float = 1):
        return self.base.calculate(x)


if __name__ == "__main__":
    from utils_minan.TimeCalculator import time_calculator
    import matplotlib.pyplot as plt
    from tqdm import tqdm

    """
    一、 UnaryPolynomial 测试
    1. calculate test
    print(UnaryPolynomial([1, 1, 1]).calculate(2))  # 1 + x + x^2|x=2 =7
    print(UnaryPolynomial([1, 1, 1]).calculate(3))  # 1 + x + x^2|x=3 =13
    print(UnaryPolynomial([1, 2, 1]).calculate(3))  # 1 + 2x + x^2|x=3 =16
    print(UnaryPolynomial([1, 2, 3]).calculate(3))  # 1 + 2x + 3x^2|x=3 =34
    print(UnaryPolynomial([1, 2, 3, 1]).calculate(3))  # 1 + 2x + 3x^2|x=3 =61
    print(UnaryPolynomial([1, 2, 3, 2]).calculate(3))  # 1 + 2x + 3x^2|x=3 =88
    2. efficient test
    time_calculator.st("test")
    time_calculator.st("my")
    for i in range(1000000):
        print(UnaryPolynomial([1, 2, 3, 2, 2, 2, 2, 2]).calculate(3))  # 1 + 2x + 3x^2|x=3 =88
    time_calculator.ed("my")

    time_calculator.st("dir")
    for i in range(1000000):
        # UnaryPolynomial([1, 2, 3, 2]).calculate(3)  # 1 + 2x + 3x^2|x=3 =88
        # y = 1 + 2 * 3. + 3 * math.pow(3., 2) + 2 * math.pow(3., 3)  # almost 4 times
        # y = 1 + 2 * 3. + 3 * math.pow(3., 2) + 2 * math.pow(3., 3) + 2 * math.pow(3., 4)  # almost 3 times
        y = 1 + 2 * 3. + 3 * math.pow(3., 2) + 2 * math.pow(3., 3) + 2 * math.pow(3., 4) \
            + 2 * math.pow(3., 5) + 2 * math.pow(3., 6) + 2 * math.pow(3., 7)  # almost 2 times
    time_calculator.ed("dir")

    time_calculator.ed("test")
    """
    """
    二、 Distribution 测试
    #d = UniformDistribution([0, 1], [0, 2])
    d = NormalDistribution([1,1,5,0,0,0,0,2], [1])
    s = []
    eval_times = 20000
    n = eval_times // 40
    bar = tqdm(eval_times)
    for i in range(eval_times):
        s.append(d.sample(2))
        if i % 50 == 49:
            bar.update(50)
    p, x = np.histogram(s, bins=n)
    x = x[:-1] + (x[1] - x[0]) / 2
    f = UnivariateSpline(x, p, s=n)
    bar.close()
    plt.plot(x, f(x))
    plt.show()
    """



from elsp_env_manager.base.constants import *


class MID(object):

    def __init__(self, _id: str):
        self._id = _id
        buff = _id.split("_")
        self.type = buff[0]
        self.id_level = 0
        for i in range(1, len(buff)):
            self.__setattr__("l{}".format(i), buff[i])
            self.id_level += 1
        assert self.id_level >= 1, _id

    def next_production_id(self):
        assert self.type == SUBSTANCE_ID_PREFIX
        assert self.id_level == 2
        return MID(self.get_id(self.id_level - 1) + '_{}'.format(int(self.get_level_value(2)) + 1))

    def get_level_value(self, lev):
        assert lev <= self.id_level
        return self.__getattribute__("l{}".format(lev))

    def get_id(self, lev=None):
        if lev is None or lev > self.id_level:
            return self._id
        else:
            _id = self.type
            for i in range(1, lev + 1):
                _id += "_{}".format(self.get_level_value(i))
            return _id

    def upper(self):
        return MID(self.get_id(self.id_level - 1))

    def same_level_id_of_type(self, _type):
        return MID(self._id.replace(self.type, _type))

    def __eq__(self, other):
        if not isinstance(other, MID):
            return False
        return self._id == other._id

    def __hash__(self):
        return self._id.__hash__()

    def __str__(self):
        return self._id



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import io

from setuptools import setup, find_packages

#############################################
# File Name: setup.py
# Author: Meeea-914
# Mail: 1028634300@qq.com
# Created Time: 2020-12-6 9:05:34
#############################################
VERSION = '0.0.33'
with io.open("README.md", encoding='utf-8') as f:
    long_description = f.read()
install_requires = open("requirements.txt").readlines()

setup(
    name="elsp_env",
    version=VERSION,
    keywords=("pip", "gymtool", "envtool", "elsptool", "elsp"),
    description="gym env elsp tool",
    long_description=long_description,
    license="MIT Licence",

    url="https://github.com/Meeea-914/elsp_env",
    author="Meeea-914",
    author_email="1028634300@qq.com",

    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=install_requires,
)



from abc import ABC, abstractmethod

from elsp_env_manager.base import MID
from elsp_env_manager.base.constants import *
from elsp_env_manager.base.model_manager import load_model
from elsp_env_manager.process_units.process_units import ProcessUnit, ProductionUnit, StorageUnit, SchedulingUnit, \
    SaleUnit, CustomerUnit, ProductStorageUnit


class ProcessLine(ProcessUnit, ABC):
    process_line = {}

    def reset(self):
        for u in self.process_unit_list:
            u.reset()

    def __init__(self, m_id: MID, model_json, env):
        super().__init__(m_id, model_json, env)
        self.process_unit_list = []

    def last_unit(self):
        return self.process_unit_list[-1] if not isinstance(self.process_unit_list[-1], ProcessLine) else self.process_unit_list[-1].last_unit()

    @abstractmethod
    def process(self, *args):
        pass

    @abstractmethod
    def set_latter_unit(self, unit):
        pass

    @abstractmethod
    def connect_units(self):
        pass

    @staticmethod
    def load(json_or_json_path, env):
        if isinstance(json_or_json_path, dict):
            json = json_or_json_path
        else:
            json = load_model(json_or_json_path)
        for l_id, l in json.items():
            # print(l_id, l)
            m_id = MID("ProcessLine_{}".format(l_id))
            if l["type"] == 'serial':
                ProcessLine.process_line[m_id] = SerialProcessLine(m_id, l, env)
            elif l["type"] == 'parallel':
                ProcessLine.process_line[m_id] = ParallelProcessLine(m_id, l, env)
            else:
                raise NotImplementedError
        for line_id, line in ProcessLine.process_line.items():
            line: ProcessLine
            line.connect_units()
        pass

    @staticmethod
    def unit(unit_type, unit_id):
        if unit_type == "product":
            unit_id_str = PRODUCTION_UNITS_ID_PREFIX + "_{}".format(unit_id)
            unit_id = MID(unit_id_str)
            unit = ProductionUnit.product_units[unit_id]
        elif unit_type == "storage":
            unit_id_str = STORAGE_UNITS_ID_PREFIX + "_{}".format(unit_id)
            unit_id = MID(unit_id_str)
            unit = StorageUnit.storage_units[unit_id]
        elif unit_type == "process_line":
            unit_id_str = PROCESS_LINES_ID_PREFIX + "_{}".format(unit_id)
            unit_id = MID(unit_id_str)
            unit = ProcessLine.process_line[unit_id]
        elif unit_type == "schedule":
            unit_id_str = SCHEDULING_UNITS_ID_PREFIX + "_{}".format(unit_id)
            unit_id = MID(unit_id_str)
            unit = SchedulingUnit.scheduling_units[unit_id]
        elif unit_type == "sale":
            unit_id_str = SALE_UNITS_ID_PREFIX + "_{}".format(unit_id)
            unit_id = MID(unit_id_str)
            unit = SaleUnit.sale_units[unit_id]
        elif unit_type == "customer":
            unit_id_str = CUSTOMER_UNITS_ID_PREFIX + "_{}".format(unit_id)
            unit_id = MID(unit_id_str)
            unit = CustomerUnit.customer_units[unit_id]
        else:
            raise NotImplementedError
        return unit


class SerialProcessLine(ProcessLine):

    def connect_units(self):
        former_unit = None
        for unit_json in self.model_json["list"]:
            unit_type: str = unit_json["type"]
            unit_id: str = unit_json["m_id"]
            unit = ProcessLine.unit(unit_type, unit_id)
            if isinstance(former_unit, ProcessUnit):
                former_unit.set_latter_unit(unit)
            former_unit = unit
            self.process_unit_list.append(unit)

    def __init__(self, m_id: MID, model_json, env):
        super().__init__(m_id, model_json, env)

    def set_latter_unit(self, unit):
        self.process_unit_list[-1].set_latter_unit(unit)

    def process(self, *args):
        self.process_unit_list[0].process(*args)
        pass


class ParallelCollectionUnit(ProcessUnit):

    def reset(self):
        pass

    def __init__(self, ppl, env):
        ppl: ParallelProcessLine
        super().__init__(ppl.m_id.same_level_id_of_type(PARALLEL_COLLECTION_UNITS_ID_PREFIX), {}, env)
        self.collection = {}
        self.current_unit = None
        self.ppl = ppl

    def process(self, *args):
        assert self.current_unit is not None
        assert len(args) == 1
        """if isinstance(self.current_unit, ProductStorageUnit):
            return"""
        self.collection[self.current_unit] = args[0]
        if self.current_unit == self.ppl.process_unit_list[-1]:
            self.get_latter_unit().process((self.collection, self.ppl.process_unit_list))

    def clear(self):
        self.collection.clear()


class ParallelProcessLine(ProcessLine):

    def set_latter_unit(self, unit):
        super().set_latter_unit(unit)
        self.collection_unit.set_latter_unit(unit)

    def get_latter_unit(self):
        return self.collection_unit.get_latter_unit()

    def connect_units(self):
        for unit_json in self.model_json["list"]:
            unit_type: str = unit_json["type"]
            unit_id: str = unit_json["m_id"]
            unit = ProcessLine.unit(unit_type, unit_id)
            unit.set_latter_unit(self.collection_unit)
            self.process_unit_list.append(unit)
        pass

    def __init__(self, m_id: MID, model_json, env):
        super().__init__(m_id, model_json, env)
        self.collection_unit = ParallelCollectionUnit(self, self.env)
        pass

    def process(self, *args):
        self.collection_unit.clear()
        for unit in self.process_unit_list:
            self.collection_unit.current_unit = unit
            unit.process(*args)
            pass



from multiprocessing.connection import Listener, Client, Connection
from multiprocessing import Process
from threading import Thread
from gym import Env
import gym
import time
import tqdm

ADDRESS = ('localhost', 6003)
AUTHKEY = b'123456'


class ClientEnv(object):

    def __init__(self, env: Env):
        self.done = True
        self.env = env
        self.obs = env.reset()
        self.conn = Client(ADDRESS, authkey=AUTHKEY)
        process = Process(target=self._main)
        process.start()

    def _main(self):
        while True:
            recv = self.conn.recv()
            if recv == "reset":
                obs = self.env.reset()
                self.conn.send(obs)
            elif recv == "destroy":
                # print("destroy")
                break
            else:
                res = self.env.step(recv)
                self.conn.send(res)


class EvalEnv(object):

    def __init__(self, policy_maker, envs, eval_data_getter, max_process=1):
        assert len(envs) >= max_process
        self.listener = Listener(ADDRESS, authkey=AUTHKEY)
        self.conn_list = []
        self.policy_maker = policy_maker
        self.eval_data_getter = eval_data_getter
        self.destroyed = False
        Thread(target=self._accept).start()
        self.client_env_list = []
        bar = tqdm.tqdm(total=max_process)
        bar.set_description("creating env")
        for i in range(max_process):
            client_env = ClientEnv(env=envs[i])
            bar.update(1)
        bar.close()
            # self.client_env_list.append(client_env)

    def _accept(self):
        while not self.destroyed:
            conn = self.listener.accept()
            self.conn_list.append(conn)
            # print("accepted:", len(self.conn_list))

    def eval(self, eval_times):
        bar = tqdm.tqdm(total=eval_times)
        bar.set_description("evaluating")
        eval_res_list = []
        obs_list = []
        action_list = []
        last_finished_num = 0
        for conn in self.conn_list:
            conn: Connection
            conn.send("reset")
            obs = conn.recv()
            obs_list.append(obs)
            action_list.append(0)
        while True:
            # print(1)
            if len(self.conn_list) > 0:
                p_n = len(self.conn_list)
                for i in range(p_n):
                    conn = self.conn_list[i]
                    obs = obs_list[i]
                    action = action_list[i]
                    action = self.policy_maker(action, obs)
                    action_list[i] = action
                    conn.send(action)
                for i in range(p_n):
                    conn = self.conn_list[i]
                    obs, reward, done, inf = conn.recv()
                    obs_list[i] = obs
                    if done:
                        eval_res_list.append(self.eval_data_getter(inf))
                        conn.send("reset")
                        obs = conn.recv()
                        obs_list[i] = obs
                        action_list[i] = action
                if len(eval_res_list) >= eval_times:
                    # print("Process:%d/%d" % (len(eval_res_list), eval_times))
                    bar.update(100 - last_finished_num)
                    bar.close()
                    return eval_res_list[0:100]
                else:
                    if last_finished_num < len(eval_res_list):
                        bar.update(len(eval_res_list) - last_finished_num)
                        last_finished_num = len(eval_res_list)
                        # print("Process:%d/%d" % (len(eval_res_list), eval_times))
            else:
                print("no conn")
                time.sleep(0.01)
            pass

    def destroy(self):
        p_n = len(self.conn_list)
        for i in range(p_n):
            conn = self.conn_list[i]
            conn.send("destroy")
        self.destroyed = True


if __name__ == "__main__":
    def env_maker():
        return gym.make("CartPole-v0")


    def policy_maker(last_action, obs):
        return 0


    def eval_data_getter(last_inf):
        return 0


    eval_env = EvalEnv(policy_maker, envs=[env_maker() for i in range(2)], max_process=2,
                       eval_data_getter=eval_data_getter)
    print(eval_env.eval(2))

